import unittest

import distro_info

import piupartslib.conf as conf


class ConfStdDistroTests(unittest.TestCase):
    def setUp(self):
        self.cobj = conf.Config("notimportant", {})

        debdist = distro_info.DebianDistroInfo()
        self.stable = debdist.stable()
        self.unstable = debdist.devel()
        self.oldstable = debdist.old()
        self.testing = debdist.testing()
        self.experimental = "experimental"

    def testConfStdDistroNames(self):
        self.assertEqual(self.unstable, "sid")
        self.assertEqual(self.experimental, "experimental")

    def testConfMapDistro(self):
        self.assertEqual(self.cobj._map_distro("bogus"), "unknown")

        self.assertEqual(self.cobj._map_distro(self.oldstable), "oldstable")
        self.assertEqual(self.cobj._map_distro(self.stable), "stable")
        self.assertEqual(self.cobj._map_distro(self.testing), "testing")
        self.assertEqual(self.cobj._map_distro(self.unstable), "unstable")
        self.assertEqual(self.cobj._map_distro(self.experimental), "experimental")

        self.assertEqual(self.cobj._map_distro("oldstable"), "oldstable")
        self.assertEqual(self.cobj._map_distro("stable"), "stable")
        self.assertEqual(self.cobj._map_distro("testing"), "testing")
        self.assertEqual(self.cobj._map_distro("unstable"), "unstable")
        self.assertEqual(self.cobj._map_distro("experimental"), "experimental")

    def testConfMapProposedDistro(self):
        self.assertEqual(self.cobj._map_distro("stable-proposed"), "stable")
        self.assertEqual(self.cobj._map_distro(self.stable + "-proposed"), "stable")

    def testConfMapRemainingDistros(self):
        self.assertEqual(self.cobj._map_distro("rc-buggy"), "experimental")

        self.assertEqual(self.cobj._map_distro("Debian6.0.9"), self.cobj._map_distro("squeeze"))
        self.assertEqual(self.cobj._map_distro("Debian7.4"), self.cobj._map_distro("wheezy"))
        self.assertEqual(self.cobj._map_distro("Debian8"), self.cobj._map_distro("jessie"))
        self.assertEqual(self.cobj._map_distro("Debian8.1"), self.cobj._map_distro("jessie"))
        self.assertEqual(self.cobj._map_distro("Debian9"), self.cobj._map_distro("stretch"))

    def testConfGetStdDistro(self):
        for std in ["oldstable", "stable", "testing", "unstable", "experimental"]:
            self.assertEqual(self.cobj.get_std_distro([self.__dict__[std]]), std)
            self.assertEqual(self.cobj.get_std_distro([self.__dict__[std], "unknown"]), std)
            self.assertEqual(self.cobj.get_std_distro(["unknown", self.__dict__[std]]), std)
            self.assertEqual(
                self.cobj.get_std_distro(["unknown", "unknown", self.__dict__[std]]),
                std,
            )
            self.assertEqual(
                self.cobj.get_std_distro([self.__dict__[std], "unknown", "unknown"]),
                std,
            )

        self.assertEqual(self.cobj.get_std_distro(["unknown"]), "unknown")
        self.assertEqual(self.cobj.get_std_distro(["unknown", "unknown"]), "unknown")



import datetime
import json
import os
import shutil
import tempfile
import unittest

import piupartslib.pkgsummary as pkgsummary


class PkgSummaryTests(unittest.TestCase):
    def testSummFlaginfoStateDups(self):
        finfo = pkgsummary.flaginfo
        states = sorted([y for x in finfo for y in finfo[x].states])
        nodups = sorted(list(set(states)))

        self.assertIn("successfully-tested", states)
        self.assertEqual(states, nodups)

    def testSummGetFlag(self):
        self.assertEqual("F", pkgsummary.get_flag("failed-testing"))
        self.assertEqual("X", pkgsummary.get_flag("dependency-does-not-exist"))
        self.assertEqual("P", pkgsummary.get_flag("successfully-tested"))
        self.assertEqual("W", pkgsummary.get_flag("waiting-to-be-tested"))

        with self.assertRaises(pkgsummary.SummaryException):
            pkgsummary.get_flag("bogus-state")

    def testSummWorstFlag(self):
        self.assertEqual("F", pkgsummary.worst_flag("F"))
        self.assertEqual("P", pkgsummary.worst_flag("P"))
        self.assertEqual("F", pkgsummary.worst_flag("P", "F"))
        self.assertEqual("F", pkgsummary.worst_flag("F", "F"))
        self.assertEqual("W", pkgsummary.worst_flag("W", "P"))
        self.assertEqual("F", pkgsummary.worst_flag("W", "P", "F", "X", "-"))

        with self.assertRaises(pkgsummary.SummaryException):
            pkgsummary.worst_flag("Z")


class PkgSummaryAddTests(unittest.TestCase):
    def setUp(self):
        self.summ = pkgsummary.new_summary()

    def testSummNewSumm(self):
        # Verify any parameters which are depended on downstream

        self.assertEqual("Piuparts Package Test Results Summary", self.summ["_id"])
        self.assertEqual("1.0", self.summ["_version"])
        self.assertEqual({}, self.summ["packages"])

        thedate = datetime.datetime.strptime(self.summ["_date"], "%a %b %d %H:%M:%S UTC %Y")
        self.assertIsNotNone(thedate)

    def testSummAddArgValidation(self):
        with self.assertRaises(pkgsummary.SummaryException):
            pkgsummary.add_summary(self.summ, "foodist", "foopkg", "Z", 0, "http://foo")
        with self.assertRaises(pkgsummary.SummaryException):
            pkgsummary.add_summary(self.summ, "foodist", "foopkg", "X", "bogus", "http://foo")
        with self.assertRaises(pkgsummary.SummaryException):
            pkgsummary.add_summary(self.summ, "foodist", "foopkg", "X", 1, "ittp://foo")

        pkgsummary.add_summary(self.summ, "foodist", "foopkg", "X", 1, "http://foo")

    def testSummAddArgStorageFormat(self):
        # store non-overlapping entries

        pkgsummary.add_summary(self.summ, "dist", "pkg", "X", 0, "http://foo")
        pkgsummary.add_summary(self.summ, "dist", "pkg2", "W", 1, "http://foo2")
        pkgsummary.add_summary(self.summ, "dist2", "pkg3", "P", 2, "http://foo3")
        self.assertEqual(["X", 0, "http://foo"], self.summ["packages"]["pkg"]["dist"])
        self.assertEqual(["W", 1, "http://foo2"], self.summ["packages"]["pkg2"]["dist"])
        self.assertEqual(["P", 2, "http://foo3"], self.summ["packages"]["pkg3"]["dist2"])

    def testSummAddOverwriteFlag(self):
        pkgsummary.add_summary(self.summ, "dist", "pkg", "X", 0, "http://foo")
        pkgsummary.add_summary(self.summ, "dist", "pkg", "P", 0, "http://foo2")
        self.assertEqual("X", self.summ["packages"]["pkg"]["dist"][0])
        self.assertEqual("http://foo", self.summ["packages"]["pkg"]["dist"][2])

        pkgsummary.add_summary(self.summ, "dist", "pkg", "F", 0, "http://foo3")
        self.assertEqual("F", self.summ["packages"]["pkg"]["dist"][0])
        self.assertEqual("http://foo3", self.summ["packages"]["pkg"]["dist"][2])

    def testSummAddBlockCount(self):
        pkgsummary.add_summary(self.summ, "dist", "pkg", "X", 0, "http://foo")
        pkgsummary.add_summary(self.summ, "dist", "pkg", "P", 1, "http://foo")
        self.assertEqual(1, self.summ["packages"]["pkg"]["dist"][1])

        pkgsummary.add_summary(self.summ, "dist", "pkg", "F", 2, "http://foo")
        self.assertEqual(2, self.summ["packages"]["pkg"]["dist"][1])

    def testSummAddInvalid(self):
        pkgsummary.add_summary(self.summ, "dist", "pkg", "-", 0, "http://foo")
        print(self.summ["packages"]["pkg"]["dist"][2])
        self.assertIn("http", self.summ["packages"]["pkg"]["dist"][2])

    def testSummMerge(self):
        pkgsummary.add_summary(self.summ, "dist", "pkg", "X", 0, "http://foo")

        mergesumm = pkgsummary.new_summary()

        pkgsummary.merge_summary(mergesumm, self.summ)

        self.assertEqual(mergesumm["packages"]["pkg"]["dist"], self.summ["packages"]["pkg"]["dist"])
        self.assertEqual(
            mergesumm["packages"]["pkg"]["dist"],
            mergesumm["packages"]["pkg"]["overall"],
        )


class PkgSummaryStorageTests(unittest.TestCase):
    def setUp(self):
        self.summ = pkgsummary.new_summary()
        pkgsummary.add_summary(self.summ, "dist", "pkg", "X", 0, "http://foo")

        self.tmpdir = tempfile.mkdtemp()

        self.tmpfilename = os.path.join(self.tmpdir, "foo.json")
        pkgsummary.write_summary(self.summ, self.tmpfilename)

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def testSummFileRead(self):
        summ2 = pkgsummary.read_summary(self.tmpfilename)

        self.assertEqual(self.summ, summ2)

    def testSummFileStorage(self):
        with open(self.tmpfilename, "r") as fl:
            summ2 = json.load(fl)

        self.assertEqual(self.summ, summ2)



import unittest

import piupartslib.dependencyparser


class DependencyParserTests(unittest.TestCase):
    """Tests for module dependencyparser."""

    def parse(self, str):
        parser = piupartslib.dependencyparser.DependencyParser(str)
        deps = parser.get_dependencies()
        names = []
        for dep in deps:
            names.append([])
            for simpledep in dep:
                names[-1].append(simpledep.name)
        return deps, names

    def testEmpty(self):
        deps, names = self.parse("")
        self.assertEqual(deps, [])

    def testSingle(self):
        deps, names = self.parse("foo")
        self.assertEqual(names, [["foo"]])

    def testTwo(self):
        deps, names = self.parse("foo, bar")
        self.assertEqual(names, [["foo"], ["bar"]])

    def testAlternatives(self):
        deps, names = self.parse("foo, bar | foobar")
        self.assertEqual(names, [["foo"], ["bar", "foobar"]])



#!/usr/bin/python3

# Copyright 2014 David Steele (dsteele@gmail.com)
#
# This file is part of Piuparts
#
# Piuparts is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Piuparts is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <http://www.gnu.org/licenses/>.


# Piuparts summary generation module
#
# This module is used to create exportable section and global package testing
# result summaries.
#
# The canonical location for section summaries is at
#
#     https://piuparts.debian.org/<section>/summary.json
#
# The global summary is at
#
#     https://piuparts.debian.org/summary.json
#
# Example output:
#
# summary.json
# {
#  "_comment": "Debian Piuparts Package Results - https://piuparts.debian.org/...",
#  "_date": "Wed Feb 26 01:48:43 UTC 2014",
#  "_id": "Piuparts Package Test Results Summary",
#  "_type": "source",
#  "_version": "1.0",
#  "packages": {
#   "0ad": {
#    "overall": [
#     "X",
#     0,
#     "http://localhost/piuparts/sid-fail-broken-symlinks/source/0/0ad.html"
#    ],
#    "stable": [
#     "P",
#     0,
#     "http://localhost/piuparts/wheezy/source/0/0ad.html"
#    ],
#    "unstable": [
#     "X",
#     0,
#     "http://localhost/piuparts/sid-fail-broken-symlinks/source/0/0ad.html"
#    ]
#   },
#   "0ad-data": {
#  ...
#  }
#
#
# The packages are listed by source package. E.g. "unstable" here is a
# json-section (see README_server.txt). The single character flags are
# defined below. The number is the number of packages which
# are blocked from testing due to a failed package. The URL is a human
# friendly page for inspecting the results for that package/distribution.
#
# Binary package results are combined into source package results. The 'worst'
# flag in the group is reported ("F" is worst overall).
#
# For the global summary, the packages 'worst' result across json-sections
# is used. In the case of a tie, the more-important-precedence
# section/json-section result is used.
#
# The global file also includes an 'overall' json-section, which contains
# the 'worst' result across the other json-sections.
from __future__ import print_function

import datetime
import json
import os
from collections import defaultdict, namedtuple

INVALID_URL = "invalid url"


class SummaryException(Exception):
    pass


SUMMID = "Piuparts Package Test Results Summary"
SUMMVER = "1.0"

DEFSEC = "overall"

FlagInfo = namedtuple("FlagInfo", ["word", "priority", "states"])

flaginfo = {
    "F": FlagInfo("Failed", 0, ["failed-testing"]),
    "X": FlagInfo(
        "Blocked",
        1,
        [
            "cannot-be-tested",
            "dependency-failed-testing",
            "dependency-cannot-be-tested",
            "dependency-does-not-exist",
        ],
    ),
    "W": FlagInfo(
        "Waiting",
        2,
        [
            "waiting-to-be-tested",
            "waiting-for-dependency-to-be-tested",
        ],
    ),
    "P": FlagInfo(
        "Passed",
        3,
        [
            "essential-required",
            "successfully-tested",
        ],
    ),
    "-": FlagInfo(
        "Unknown",
        4,
        [
            "does-not-exist",
            "unknown",
        ],
    ),
}

state2flg = dict([(y, x[0]) for x in flaginfo.items() for y in x[1].states])


def worst_flag(*flags):
    try:
        flag = min(*flags, key=lambda x: flaginfo[x].priority)
    except KeyError:
        raise SummaryException("Unknown flag in " + flags.__repr__())

    return flag


def get_flag(state):
    try:
        flag = state2flg[state]
    except KeyError:
        raise SummaryException("Unknown state - " + state)

    return flag


def new_summary():
    cdate_array = datetime.datetime.utcnow().ctime().split()
    utcdate = " ".join(cdate_array[:-1] + ["UTC"] + [cdate_array[-1]])

    # define the packages struct. The default should never be the one added
    dfltentry = ["-", 0, INVALID_URL]
    pkgstruct = defaultdict(lambda: defaultdict(lambda: dfltentry))

    return {
        "_id": SUMMID,
        "_version": SUMMVER,
        "_date": utcdate,
        "_comment": "Debian Piuparts Package Results - "
        "https://salsa.debian.org/debian/piuparts/raw/"
        "develop/piupartslib/pkgsummary.py",
        "_type": "source",
        "packages": pkgstruct,
    }


def add_summary(summary, rep_sec, pkg, flag, block_cnt, url):
    if flag not in flaginfo or not isinstance(block_cnt, int) or not url.startswith("http"):
        raise SummaryException("Invalid summary argument")

    pdict = summary["packages"]

    [old_flag, old_cnt, old_url] = pdict[pkg][rep_sec]
    block_cnt = max(block_cnt, old_cnt)
    if old_flag != worst_flag(old_flag, flag) or old_url == INVALID_URL:
        pdict[pkg][rep_sec] = [flag, block_cnt, url]
    else:
        pdict[pkg][rep_sec] = [old_flag, block_cnt, old_url]

    return summary


def merge_summary(gbl_summ, sec_summ):
    spdict = sec_summ["packages"]

    for pkg in spdict:
        for rep_sec in spdict[pkg]:
            flag, block_cnt, url = spdict[pkg][rep_sec]
            add_summary(gbl_summ, rep_sec, pkg, flag, block_cnt, url)
            add_summary(gbl_summ, DEFSEC, pkg, flag, block_cnt, url)

    return gbl_summ


def tooltip(summary, pkg):
    """Returns e.g. "Failed in testing and stable, blocking 5 packages"."""

    tip = ""

    pkgdict = summary["packages"]

    if pkg in pkgdict:
        flag, block_cnt, url = pkgdict[pkg][DEFSEC]

        sections = [x for x in pkgdict[pkg] if x != DEFSEC]
        applicable = [x for x in sections if pkgdict[pkg][x][0] == flag]

        tip = flaginfo[flag].word

        if len(applicable) > 2:
            tip += " in " + ", ".join(applicable[:-1]) + " and " + applicable[-1]
        elif len(applicable) == 2:
            tip += " in " + " and ".join(applicable)
        elif len(applicable) == 1:
            tip += " in " + applicable[0]

        if block_cnt:
            tip += ", blocking %d packages" % block_cnt

        tip += "."

    return tip


def write_summary(summary, fname):
    with open(fname + ".tmp", "w") as fl:
        json.dump(summary, fl, sort_keys=True, indent=1)
    os.rename(fname + ".tmp", fname)


def read_summary(fname):
    with open(fname, "r") as fl:
        result = json.load(fl)

    if result["_id"] != SUMMID or result["_version"] != SUMMVER:
        raise SummaryException("Summary JSON header mismatch")

    return result


if __name__ == "__main__":
    import sys

    # read a global summary file and return DDPO info by package
    summary = read_summary(sys.argv[1])

    for pkg in summary["packages"]:
        flag, blocked, url = summary["packages"][pkg][DEFSEC]

        print(pkg, flag, url, tooltip(summary, pkg))



# -*- coding: utf-8 -*-

# Copyright 2005 Lars Wirzenius (liw@iki.fi)
# Copyright © 2012 Andreas Beckmann (anbe@debian.org)
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <https://www.gnu.org/licenses/>


"""Parser for Debian package relationship strings

This module contains the class DependencyParser, which parses Debian
package relationship strings (e.g., the Depends header). The class
raises the DependencySyntaxError exception on syntactic errors.
The result uses SimpleDependency objects.

Lars Wirzenius <liw@iki.fi>
"""


import re


class DependencySyntaxError(Exception):
    """Syntax error in package dependency declaration"""

    def __init__(self, msg, cursor):
        self._msg = "Error: %s: %s (text at error: '%s', full text being parsed: '%s')" % (
            cursor.get_position(),
            msg,
            cursor.get_text(10),
            cursor.get_full_text(),
        )

    def __str__(self):
        return self._msg

    def __repr__(self):
        return self._msg


class _Cursor:
    """Store an input string and a movable location in it"""

    def __init__(self, myinput):
        self._input = myinput
        self._len = len(self._input)
        self._pos = 0

    def skip_whitespace(self):
        while self._pos < self._len and self._input[self._pos].isspace():
            self.mynext()

    def at_end(self):
        """Are we at the end of the input?"""
        self.skip_whitespace()
        return self._pos >= self._len

    def mynext(self):
        """Move to the next character"""
        if self._pos < self._len:
            self._pos += 1

    def get_char(self):
        """Return current character, None if at end"""
        if self._pos >= self._len:
            return None
        else:
            return self._input[self._pos]

    def get_full_text(self):
        return self._input

    def get_text(self, length):
        """Return up to length characters from the current position"""
        if self._pos >= self._len:
            return ""
        else:
            return self._input[self._pos : self._pos + length]

    def match(self, regexp):
        """Match a regular expression against the current position

        The cursor is advanced by the length of the match, if any.

        """
        m = regexp.match(self._input[self._pos :])
        if m:
            self._pos += len(m.group())
        return m

    def match_literal(self, literal):
        """Match a literal string against the current position.

        Return True and move position if there is a match, else return
        False.

        """
        if self.get_text(len(literal)) == literal:
            self._pos += len(literal)
            return True
        else:
            return False

    def get_position(self):
        """Return current position, as string"""
        return "pos %d" % self._pos


class SimpleDependency:
    """Express simple dependency towards another package"""

    def __init__(self, name, operator, version, arch):
        self.name = name
        self.operator = operator
        self.version = version
        self.arch = arch

    def __repr__(self):
        return "<DEP: %s, %s, %s, %s>" % (
            self.name,
            self.operator,
            self.version,
            self.arch,
        )


class DependencyParser:
    """Parse Debian package relationship strings

    Debian packages have a rich language for expressing their
    relationships. See the Debian Policy Manual, chapter 7 ("Declaring
    relationships between packages"). This Python module implements a
    parser for strings expressing such relationships.

    Syntax of dependency fields (Pre-Depends, Depends, Recommends,
    Suggests, Conflicts, Provides, Replaces, Enhances, Build-Depends,
    Build-Depends-Indep, Build-Conflicts, Build-Conflicts-Indep), in a
    BNF-like form:

        depends-field ::= EMPTY | dependency ("," dependency)*
        dependency ::= possible-dependency ("|" possible-dependency)*
        possible-dependency ::= package-name version-dependency?
                                arch-restriction?
        version-dependency ::= "(" relative-operator version-number ")"
        relative-operator ::= "<<" | "<=" | "=" | ">=" | ">>" | "<" | ">"
        version-number ::= epoch? upstream-version debian-revision?
        arch-restriction ::= "[" arch-name arch-name* "]" |
                              "[" "!" arch-name ("!" arch-name)* "]"
        package-name ::= alphanumeric name-char name-char* ":any"?
        epoch ::= integer ":"
        upstream-version ::= alphanumeric version-char*
            -- policy says "should start with digit", but not all packages do
        debian-revision ::= "-" debian-version-char debian-version-char*
        arch-name ::= alphanumeric alphanumeric*
        EMPTY ::= ""
        integer ::= digit digit*
        alphanumeric ::=
            "a" | "b" | "c" | "d" | "e" | "f" | "g" | "h" | "i" | "j" |
            "k" | "l" | "m" | "n" | "o" | "p" | "q" | "r" | "s" | "t" |
            "u" | "v" | "w" | "x" | "y" | "z" | digit
        digit ::= "0" | "1" | "2" | "3" | "4" | "5" | "6" | "7" | "8" | "9"
        name-char ::= alphanumeric | "+" | "-" | "." | "_"
        version-char ::= alphanumeric | "." | "+" | "-" | ":" | "~"
        debian-version-char ::= alphanumeric | "." | "+"

    White space can occur between any tokens except inside package-name,
    version-number, or arch-name. Some of the headers restrict the syntax
    somewhat, e.g., Provides does not allow version-dependency, but this is
    not included in the syntax for simplicity.

    Note: Added "_" to name-char, because some packages (type-handling
    in particular) use Provides: headers with bogus package names.

    Note: Added upper case letters to name pattern, since it some of the
    Mozilla localization packages use or used them.

    """

    def __init__(self, input_string):
        self._cursor = _Cursor(input_string)
        self._list = self._parse_dependencies()

    def get_dependencies(self):
        """Return parsed dependencies

        The result is a list of lists of SimpleDependency objects.
        Let's try that again.

        The result is a list of dependencies, corresponding to
        the comma-separated items in the dependency list. Each dependency
        is also a list, or SimpleDependency objects, representing
        alternative ways to fulfill the dependency; in other words,
        items separated by the vertical bar (|).

        For example, "foo, bar | foobar" would result in the following
        list: [[foo], [bar, foobar]].

        """
        return self._list

    def _parse_dependencies(self):
        vlist = []
        dep = self._parse_dependency()
        while dep:
            vlist.append(dep)
            self._cursor.skip_whitespace()
            if self._cursor.at_end():
                break
            if not self._cursor.match_literal(","):
                raise DependencySyntaxError("Expected comma", self._cursor)
            dep = self._parse_dependency()
        return vlist

    def _parse_dependency(self):
        vlist = []
        dep = self._parse_possible_dependency()
        while dep:
            vlist.append(dep)
            self._cursor.skip_whitespace()
            if not self._cursor.match_literal("|"):
                break
            dep = self._parse_possible_dependency()
        return vlist

    def _parse_possible_dependency(self):
        name = self._parse_package_name()
        if not name:
            return None
        (op, version) = self._parse_version_dependency()
        arch = self._parse_arch_restriction()
        return SimpleDependency(name, op, version, arch)

    _name_pat = re.compile(r"[a-zA-Z0-9][a-zA-Z0-9+._-]+")
    # The MultiArch spec supports an ":any" modifier. Loosen the
    # accepted arch's, to avoid crashing.
    _any_suffix_pat = re.compile(r":[a-zA-Z0-9-]+")

    def _parse_package_name(self):
        self._cursor.skip_whitespace()
        if self._cursor.at_end():
            return None
        m = self._cursor.match(self._name_pat)
        if not m:
            raise DependencySyntaxError("Expected a package name", self._cursor)
        if self._cursor.match(self._any_suffix_pat):
            pass
        return m.group()

    _op_pat = re.compile(r"(<<|<=|=|>=|>>|<(?![<=])|>(?![>=]))")
    _version_pat = re.compile(
        r"(?P<epoch>\d+:)?" + r"(?P<upstream>[a-zA-Z0-9+][a-zA-Z0-9.+:~-]*)" + r"(?P<debian>-[a-zA-Z0-9.+]+)?"
    )

    def _parse_version_dependency(self):
        self._cursor.skip_whitespace()
        if self._cursor.get_char() == "(":
            self._cursor.mynext()

            self._cursor.skip_whitespace()
            opm = self._cursor.match(self._op_pat)
            if not opm:
                raise DependencySyntaxError("Expected a version relation " + "operator", self._cursor)
            operator = opm.group()
            if operator == "<":
                operator = "<="
            elif operator == ">":
                operator = ">="

            self._cursor.skip_whitespace()
            verm = self._cursor.match(self._version_pat)
            if not verm:
                raise DependencySyntaxError("Expected a version number", self._cursor)

            self._cursor.skip_whitespace()
            if self._cursor.get_char() != ")":
                raise DependencySyntaxError("Expected ')'", self._cursor)
            self._cursor.mynext()

            return opm.group(), verm.group()
        else:
            return None, None

    _arch_pat = re.compile(r"!?[a-zA-Z0-9-]+")

    def _parse_arch_restriction(self):
        self._cursor.skip_whitespace()
        if self._cursor.get_char() == "[":
            self.mynext()

            vlist = []
            while True:
                self._cursor.skip_whitespace()
                if self._cursor.get_char() == "]":
                    self._cursor.mynext()
                    break
                m = self._cursor.match(self._arch_pat)
                if not m:
                    raise DependencySyntaxError("Expected architecture name", self._cursor)
                vlist.append(m.group())

            return vlist
        else:
            return None


# vi:set et ts=4 sw=4 :



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright 2005 Lars Wirzenius (liw@iki.fi)
# Copyright © 2011-2017 Andreas Beckmann (anbe@debian.org)
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <https://www.gnu.org/licenses/>


"""Distributed piuparts processing, master program

Lars Wirzenius <liw@iki.fi>
"""


import fcntl
import logging
import os
import random
import sys
import time
from urllib.error import URLError

import piupartslib.conf
import piupartslib.packagesdb
from piupartslib.conf import MissingSection
from piupartslib.packagesdb import LogfileExists

CONFIG_FILE = "/etc/piuparts/piuparts.conf"
DISTRO_CONFIG_FILE = "/etc/piuparts/distros.conf"


log_handler = None


def setup_logging(log_level, log_file_name):
    logger = logging.getLogger()

    global log_handler
    logger.removeHandler(log_handler)

    if log_file_name:
        log_handler = logging.FileHandler(log_file_name)
    else:
        log_handler = logging.StreamHandler(sys.stderr)

    logger.addHandler(log_handler)
    logger.setLevel(log_level)


def timestamp():
    return time.strftime("[%Y-%m-%d %H:%M:%S]")


class Config(piupartslib.conf.Config):
    def __init__(self, section="master", defaults_section=None):
        piupartslib.conf.Config.__init__(
            self,
            section,
            {
                "log-file": None,
                "master-directory": ".",
                "proxy": None,
                "mirror": None,
                "distro": None,
                "area": None,
                "arch": None,
                "upgrade-test-distros": None,
                "depends-sections": None,
            },
            defaults_section=defaults_section,
        )


class CommandSyntaxError(Exception):
    def __init__(self, msg):
        self.args = (msg,)


class ProtocolError(Exception):
    def __init__(self):
        self.args = ("EOF, missing space in long part, or other protocol error",)


class Protocol:
    def __init__(self, myinput, output):
        self._input = myinput
        self._output = output

    def _readline(self):
        line = self._input.readline()
        logging.debug(">> " + line.rstrip())
        return line

    def _writeline(self, line):
        logging.debug("<< " + line)
        self._output.write(line + "\n")
        self._output.flush()

    def _short_response(self, *words):
        self._writeline(" ".join(words))

    def _read_long_part(self):
        lines = []
        while True:
            line = self._input.readline()
            if not line:
                raise ProtocolError()
            if line == ".\n":
                break
            if line[0] != " ":
                raise ProtocolError()
            lines.append(line[1:])
        return "".join(lines)


class Master(Protocol):
    def __init__(self, myinput, output):
        Protocol.__init__(self, myinput, output)
        self._commands = {
            "section": self._switch_section,
            "recycle": self._recycle,
            "idle": self._idle,
            "status": self._status,
            "reserve": self._reserve,
            "unreserve": self._unreserve,
            "pass": self._pass,
            "fail": self._fail,
            "untestable": self._untestable,
            # debug commands, unstable and undocumented interface
            "_state": self._state,
            "_depends": self._depends,
            "_recursive-depends": self._recursive_depends,
            "_depcycle": self._depcycle,
            "_list": self._list,
        }
        self._section = None
        self._lock = None
        self._writeline("hello")

    def _init_section(self, section):
        if self._lock:
            self._lock.close()

        # clear all settings from a previous section and set defaults
        self._section = None
        self._lock = None
        self._recycle_mode = False
        self._idle_mode = False
        self._idle_stamp = os.path.join(section, "idle.stamp")
        self._package_databases = None
        self._binary_db = None

        config = Config(section=section, defaults_section="global")
        try:
            config.read(CONFIG_FILE)
        except MissingSection:
            return False

        if not os.path.exists(section):
            os.makedirs(section)

        self._lock = open(os.path.join(section, "master.lock"), "w")
        try:
            fcntl.flock(self._lock, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except IOError:
            return False

        self._section = section

        logging.debug(timestamp() + " switching logfile")
        logfile = config["log-file"] or os.path.join(section, "master.log")
        setup_logging(logging.DEBUG, logfile)
        logging.debug(timestamp() + " connected")

        # start with a dummy _binary_db (without Packages file), sufficient
        # for submitting finished logs
        self._binary_db = piupartslib.packagesdb.PackagesDB(prefix=section)

        return True

    def _init_db(self):
        if self._package_databases is not None:
            return

        self._package_databases = {}
        self._load_package_database(self._section)
        self._binary_db = self._package_databases[self._section]

    def _load_package_database(self, section):
        if section in self._package_databases:
            return

        config = Config(section=section, defaults_section="global")
        config.read(CONFIG_FILE)
        distro_config = piupartslib.conf.DistroConfig(DISTRO_CONFIG_FILE, config["mirror"])
        db = piupartslib.packagesdb.PackagesDB(prefix=section)
        if self._recycle_mode and self._section == section:
            db.enable_recycling()
        self._package_databases[section] = db
        if config["depends-sections"]:
            deps = config["depends-sections"].split()
            for dep in deps:
                self._load_package_database(dep)
            db.set_dependency_databases([self._package_databases[dep] for dep in deps])
        db.load_packages_urls(
            distro_config.get_packages_urls(config.get_distro(), config.get_area(), config.get_arch())
        )
        if config.get_distro() != config.get_final_distro():
            # take version numbers (or None) from final distro
            db.load_alternate_versions_from_packages_urls(
                distro_config.get_packages_urls(config.get_final_distro(), config.get_area(), config.get_arch())
            )

    def _clear_idle(self):
        if self._idle_mode:
            self._idle_mode = False
            if os.path.exists(self._idle_stamp):
                os.unlink(self._idle_stamp)

    def _set_idle(self):
        if not self._idle_mode:
            self._idle_mode = True
            open(self._idle_stamp, "w").close()
            os.utime(self._idle_stamp, (-1, self._binary_db._stamp))

    def _get_idle_status(self):
        """Returns number of seconds a cached idle status is still valid, or 0 if not known to be idle."""
        if not os.path.exists(self._idle_stamp):
            return 0
        stamp_mtime = os.path.getmtime(self._idle_stamp)
        ttl = stamp_mtime + 3600 - time.time()
        if ttl <= 0:
            return 0  # stamp expired
        if stamp_mtime < self._binary_db.get_mtime():
            return 0  # stamp outdated
        return ttl + random.randrange(120)

    def do_transaction(self):
        line = self._readline()
        if line:
            parts = line.split()
            if len(parts) > 0:
                command = parts[0]
                args = parts[1:]
                if self._section is None and command != "section":
                    raise CommandSyntaxError("Expected 'section' command, got %s" % command)
                if command in self._commands:
                    self._commands[command](command, args)
                    return True
                else:
                    raise CommandSyntaxError("Unknown command %s" % command)
        return False

    def _check_args(self, count, command, args):
        if len(args) != count:
            raise CommandSyntaxError("Need exactly %d args: %s %s" % (count, command, " ".join(args)))

    def dump_pkgs(self):
        for st in self._binary_db.get_states():
            for name in self._binary_db.get_pkg_names_in_state(st):
                logging.debug("%s : %s\n" % (st, name))

    def _switch_section(self, command, args):
        self._check_args(1, command, args)
        if self._init_section(args[0]):
            self._short_response("ok")
        elif self._lock is None:
            # unknown section
            self._short_response("error")
        else:
            self._short_response("busy")

    def _recycle(self, command, args):
        self._check_args(0, command, args)
        if self._binary_db.enable_recycling():
            self._idle_stamp = os.path.join(self._section, "recycle.stamp")
            self._recycle_mode = True
            self._short_response("ok")
        else:
            self._short_response("error")

    def _idle(self, command, args):
        self._check_args(0, command, args)
        self._short_response("ok", "%d" % self._get_idle_status())

    def _status(self, command, args):
        self._check_args(0, command, args)
        self._init_db()
        stats = ""
        if self._binary_db._recycle_mode:
            stats += "(recycle) "
        total = 0
        for state in self._binary_db.get_active_states():
            count = len(self._binary_db.get_pkg_names_in_state(state))
            total += count
            stats += "%s=%d " % (state, count)
        stats += "total=%d" % total
        self._short_response("ok", stats)

    def _reserve(self, command, args):
        self._check_args(0, command, args)
        self._init_db()
        package = self._binary_db.reserve_package()
        if package is None:
            self._set_idle()
            self._short_response("error")
        else:
            self._clear_idle()
            self._short_response("ok", package.name(), package.test_versions())

    def _unreserve(self, command, args):
        self._check_args(2, command, args)
        self._binary_db.unreserve_package(args[0], args[1])
        self._short_response("ok")

    def _pass(self, command, args):
        self._check_args(2, command, args)
        log = self._read_long_part()
        try:
            self._binary_db.pass_package(args[0], args[1], log)
        except LogfileExists:
            logging.info("Ignoring duplicate submission: %s %s %s" % ("pass", args[0], args[1]))
        self._short_response("ok")

    def _fail(self, command, args):
        self._check_args(2, command, args)
        log = self._read_long_part()
        try:
            self._binary_db.fail_package(args[0], args[1], log)
        except LogfileExists:
            logging.info("Ignoring duplicate submission: %s %s %s" % ("fail", args[0], args[1]))
        self._short_response("ok")

    def _untestable(self, command, args):
        self._check_args(2, command, args)
        log = self._read_long_part()
        try:
            self._binary_db.make_package_untestable(args[0], args[1], log)
        except LogfileExists:
            logging.info("Ignoring duplicate submission: %s %s %s" % ("untestable", args[0], args[1]))
        self._short_response("ok")

    # debug command
    def _state(self, command, args):
        self._check_args(1, command, args)
        self._short_response(
            "ok",
            self._binary_db.get_package_state(args[0]),
            self._binary_db.get_test_versions(args[0]),
        )

    # debug command
    def _depends(self, command, args):
        self._check_args(1, command, args)
        if self._binary_db.has_package(args[0]):
            package = self._binary_db.get_package(args[0])
            self._short_response("ok", *package.dependencies())
        else:
            self._short_response("error")

    # debug command
    def _recursive_depends(self, command, args):
        self._check_args(1, command, args)
        if self._binary_db.has_package(args[0]):
            package = self._binary_db.get_package(args[0])
            self._short_response("ok", *self._binary_db._get_recursive_dependencies(package))
        else:
            self._short_response("error")

    # debug command
    def _depcycle(self, command, args):
        self._check_args(1, command, args)
        if self._binary_db.has_package(args[0]):
            self._short_response("ok", *self._binary_db._get_dependency_cycle(args[0]))
        else:
            self._short_response("error")

    # debug command
    def _list(self, command, args):
        self._check_args(1, command, args)
        if args[0] in self._binary_db.get_states():
            self._short_response("ok", *self._binary_db.get_pkg_names_in_state(args[0]))
        else:
            self._short_response("error")


def main():
    setup_logging(logging.INFO, None)
    global_config = Config(section="global")
    global_config.read(CONFIG_FILE)
    if global_config["proxy"]:
        os.environ["http_proxy"] = global_config["proxy"]
    master_directory = global_config["master-directory"]

    if not os.path.exists(master_directory):
        os.makedirs(master_directory)

    os.chdir(master_directory)

    m = Master(sys.stdin, sys.stdout)
    try:
        while m.do_transaction():
            pass
    except URLError as e:
        logging.error("ABORT: URLError: " + str(e.reason))
    except BrokenPipeError:
        logging.error("ABORT: BrokenPipeError")
        logging.debug(timestamp() + " disconnected")
        # https://docs.python.org/3/library/signal.html#note-on-sigpipe
        devnull = os.open(os.devnull, os.O_WRONLY)
        os.dup2(devnull, sys.stdout.fileno())
        sys.exit(1)

    logging.debug(timestamp() + " disconnected")


if __name__ == "__main__":
    main()

# vi:set et ts=4 sw=4 :



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = "Piuparts"
copyright = ""
author = ""

# The short X.Y version
version = ""
# The full version, including alpha/beta/rc tags
release = "0.98"


# source_suffix = ['.rst', '.md']
source_suffix = ".txt"

# The master toctree document.
master_doc = "index"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "alabaster"

# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "piupartsdoc"


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        master_doc,
        "piuparts_slave_run",
        "start up piuparts-slave instance(s) in screen\n",
        "",
        8,
    )
]



import os
import shutil
import unittest
from unittest.mock import patch

import piuparts
from piuparts import is_broken_symlink


class DefaultsFactoryTests(unittest.TestCase):
    def setUp(self):
        self.df = piuparts.DefaultsFactory()
        piuparts.settings = piuparts.Settings()

    def test_new_defaults_return_debian_defaults(self):
        # mock the guess_flavor function as it runs lsb_release in a subprocess
        with patch.object(self.df, "guess_flavor", return_value="debian") as guess_flavor_mock:
            defaults = self.df.new_defaults()
            guess_flavor_mock.assert_called_once()

        self.assertEqual(defaults.get_keyring(), "/usr/share/keyrings/debian-archive-keyring.gpg")
        self.assertEqual(
            defaults.get_components(),
            ["main", "contrib", "non-free", "non-free-firmware"],
        )
        self.assertEqual(
            defaults.get_mirror(),
            [
                (
                    "http://deb.debian.org/debian",
                    ["main", "contrib", "non-free", "non-free-firmware"],
                )
            ],
        )
        self.assertEqual(defaults.get_distribution(), ["sid"])

    def test_new_defaults_return_ubuntu_defaults(self):
        with patch.object(self.df, "guess_flavor", return_value="ubuntu") as guess_flavor_mock:
            defaults = self.df.new_defaults()
            guess_flavor_mock.assert_called_once()

        self.assertEqual(defaults.get_keyring(), "/usr/share/keyrings/ubuntu-archive-keyring.gpg")
        self.assertEqual(defaults.get_components(), ["main", "universe", "restricted", "multiverse"])
        self.assertEqual(
            defaults.get_mirror(),
            [
                (
                    "http://archive.ubuntu.com/ubuntu",
                    ["main", "universe", "restricted", "multiverse"],
                )
            ],
        )

    def test_new_defaults_panics_with_unknown_flavor(self):
        with (
            patch.object(self.df, "guess_flavor", return_value="centos") as guess_flavor_mock,
            patch.object(piuparts, "panic", side_effect=SystemExit) as panic_mock,
        ):
            with self.assertRaises(SystemExit):
                self.df.new_defaults()

            guess_flavor_mock.assert_called_once()
            panic_mock.assert_called_once()


class IsBrokenSymlinkTests(unittest.TestCase):
    testdir = "is-broken-symlink-testdir"

    def symlink(self, target, name):
        pathname = os.path.join(self.testdir, name)
        os.symlink(target, pathname)
        self.symlinks.append(pathname)

    def setUp(self):
        self.symlinks = []
        os.mkdir(self.testdir)
        self.symlink("notexist", "relative-broken")
        self.symlink("relative-broken", "relative-broken-to-symlink")
        self.symlink(".", "relative-works")
        self.symlink("relative-works", "relative-works-to-symlink")
        self.symlink("/etc", "absolute-broken")
        self.symlink("absolute-broken", "absolute-broken-to-symlink")
        self.symlink("/", "absolute-works")
        self.symlink("/absolute-works", "absolute-works-to-symlink")
        os.mkdir(os.path.join(self.testdir, "dir"))
        self.symlink("dir", "dir-link")
        os.mkdir(os.path.join(self.testdir, "dir/subdir"))
        self.symlink("subdir", "dir/subdir-link")
        self.symlink("notexist/", "trailing-slash-broken")
        self.symlink("dir/", "trailing-slash-works")
        self.symlink("selfloop", "selfloop")
        self.symlink("/absolute-selfloop", "absolute-selfloop")
        self.symlink("../dir/selfloop", "dir/selfloop")
        self.symlink("../dir-link/selfloop", "dir/selfloop1")
        self.symlink("../../dir/subdir/selfloop", "dir/subdir/selfloop")
        self.symlink("../../dir-link/subdir/selfloop", "dir/subdir/selfloop1")
        self.symlink("../../link/subdir-link/selfloop", "dir/subdir/selfloop2")
        self.symlink("../../dir-link/subdir-link/selfloop", "dir/subdir/selfloop3")
        self.symlink("explode/bomb", "explode")

    def tearDown(self):
        shutil.rmtree(self.testdir)

    def testRelativeBroken(self):
        self.assertTrue(is_broken_symlink(self.testdir, self.testdir, "relative-broken"))

    def testRelativeBrokenToSymlink(self):
        self.assertTrue(is_broken_symlink(self.testdir, self.testdir, "relative-broken-to-symlink"))

    def testAbsoluteBroken(self):
        self.assertTrue(is_broken_symlink(self.testdir, self.testdir, "absolute-broken"))

    def testAbsoluteBrokenToSymlink(self):
        self.assertTrue(is_broken_symlink(self.testdir, self.testdir, "absolute-broken-to-symlink"))

    def testTrailingSlashBroken(self):
        self.assertTrue(is_broken_symlink(self.testdir, self.testdir, "trailing-slash-broken"))

    def testSelfLoopBroken(self):
        self.assertTrue(is_broken_symlink(self.testdir, self.testdir, "selfloop"))

    def testExpandingSelfLoopBroken(self):
        self.assertTrue(is_broken_symlink(self.testdir, self.testdir, "explode"))

    def testAbsoluteSelfLoopBroken(self):
        self.assertTrue(is_broken_symlink(self.testdir, self.testdir, "absolute-selfloop"))

    def testSubdirSelfLoopBroken(self):
        self.assertTrue(is_broken_symlink(self.testdir, self.testdir, "dir/selfloop"))
        self.assertTrue(is_broken_symlink(self.testdir, self.testdir, "dir/selfloop1"))
        self.assertTrue(is_broken_symlink(self.testdir, self.testdir, "dir/subdir/selfloop"))
        self.assertTrue(is_broken_symlink(self.testdir, self.testdir, "dir/subdir/selfloop1"))
        self.assertTrue(is_broken_symlink(self.testdir, self.testdir, "dir/subdir/selfloop2"))
        self.assertTrue(is_broken_symlink(self.testdir, self.testdir, "dir/subdir/selfloop3"))

    def testRelativeWorks(self):
        self.assertFalse(is_broken_symlink(self.testdir, self.testdir, "relative-works"))

    def testRelativeWorksToSymlink(self):
        self.assertFalse(is_broken_symlink(self.testdir, self.testdir, "relative-works-to-symlink"))

    def testAbsoluteWorks(self):
        self.assertFalse(is_broken_symlink(self.testdir, self.testdir, "absolute-works"))

    def testAbsoluteWorksToSymlink(self):
        self.assertFalse(is_broken_symlink(self.testdir, self.testdir, "absolute-works-to-symlink"))

    def testTrailingSlashWorks(self):
        self.assertFalse(is_broken_symlink(self.testdir, self.testdir, "trailing-slash-works"))

    def testMultiLevelNestedSymlinks(self):
        # target/first-link -> ../target/second-link -> ../target

        os.mkdir(os.path.join(self.testdir, "target"))
        self.symlink("../target", "target/second-link")
        self.symlink("../target/second-link", "target/first-link")
        self.assertFalse(is_broken_symlink(self.testdir, self.testdir, "target/first-link"))

    def testMultiLevelNestedAbsoluteSymlinks(self):
        # first-link -> /second-link/final-target
        # second-link -> /target-dir

        os.mkdir(os.path.join(self.testdir, "final-dir"))
        os.mkdir(os.path.join(self.testdir, "final-dir/final-target"))
        self.symlink("/second-link/final-target", "first-link")
        self.symlink("/final-dir", "second-link")
        self.assertFalse(is_broken_symlink(self.testdir, self.testdir, "first-link"))



#!/usr/bin/python3
# -*- coding: utf-8 -*-

# Copyright 2013 David Steele (dsteele@gmail.com)
# Copyright © 2014 Andreas Beckmann (anbe@debian.org)
# Copyright © 2017-2018 Holger Levsen (holger@layer-acht.org)
#
# This file is part of Piuparts
#
# Piuparts is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Piuparts is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <http://www.gnu.org/licenses/>.


import argparse
import fcntl
import logging
import os
import sys
import time
from collections import deque

import piupartslib.conf
from piupartslib.conf import MissingSection
from piupartslib.dwke import BUG_EXT, KPR_EXT, LOG_EXT, clean_cache_files, create_problem_list, get_file_dict, make_kprs

CONFIG_FILE = "/etc/piuparts/piuparts.conf"
KPR_DIRS = ("pass", "bugged", "affected", "fail", "untestable")


class Busy(Exception):
    def __init__(self):
        self.args = ("section is locked by another process",)


class WKE_Config(piupartslib.conf.Config):
    """Configuration parameters for Well Known Errors"""

    def __init__(self, section="global", defaults_section=None):
        self.section = section
        piupartslib.conf.Config.__init__(
            self,
            section,
            {
                "sections": "report",
                "master-directory": ".",
                "known-problem-directory": "@sharedir@/piuparts/known_problems",
                "exclude-known-problems": None,
            },
            defaults_section=defaults_section,
        )


def setup_logging(log_level):
    logger = logging.getLogger()
    logger.setLevel(log_level)
    handler = logging.StreamHandler(sys.stdout)
    logger.addHandler(handler)


def process_section(section, config, problem_list, recheck=False, recheck_failed=False, pkgsdb=None):
    """Update .bug and .kpr files for logs in this section"""

    sectiondir = os.path.join(config["master-directory"], section)
    workdirs = [os.path.join(sectiondir, x) for x in KPR_DIRS]

    if not os.access(sectiondir, os.F_OK):
        raise MissingSection("", section)

    with open(os.path.join(sectiondir, "master.lock"), "w") as lock:
        try:
            fcntl.flock(lock, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except IOError:
            raise Busy()

        [os.mkdir(x) for x in workdirs if not os.path.exists(x)]

        logdict = get_file_dict(workdirs, LOG_EXT)
        kprdict = get_file_dict(workdirs, KPR_EXT)
        bugdict = get_file_dict(workdirs, BUG_EXT)

        del_cnt = clean_cache_files(logdict, kprdict, recheck, recheck_failed)
        clean_cache_files(logdict, bugdict, skipnewer=True)

        kprdict = get_file_dict(workdirs, KPR_EXT)

        section_config = WKE_Config(section=section, defaults_section="global")
        section_config.read(CONFIG_FILE)
        if section_config["exclude-known-problems"]:
            excluded = section_config["exclude-known-problems"].split()
            problem_list = [p for p in problem_list if p.name not in excluded]

        add_cnt = make_kprs(logdict, kprdict, problem_list)

        return (del_cnt, add_cnt)


def detect_well_known_errors(sections, config, problem_list, recheck, recheck_failed):
    total_del = 0
    total_add = 0
    todo = deque([(s, 0) for s in sections])
    while len(todo):
        (section, next_try) = todo.popleft()
        now = time.time()
        if now < next_try:
            # sleeping, section has been tried recently
            time.sleep(max(30, next_try - now) + 30)
        try:
            (del_cnt, add_cnt) = process_section(section, config, problem_list, recheck, recheck_failed)
            total_del += del_cnt
            total_add += add_cnt
            current_time = time.strftime("%a %b %2d %H:%M:%S %Z %Y", time.localtime())
            if del_cnt == 0 and add_cnt == 0:
                pass  # nothing new
            else:
                logging.info(
                    "%s - %s: parsed logfiles: %d removed, %d added" % (current_time, section, del_cnt, add_cnt)
                )
        except Busy:
            # section is busy
            todo.append((section, time.time() + 300))
        except MissingSection:
            pass

    current_time = time.strftime("%a %b %2d %H:%M:%S %Z %Y", time.localtime())
    logging.info("%s - total parsed logfiles: %d removed, %d added" % (current_time, total_del, total_add))
    logging.info("")


if __name__ == "__main__":
    setup_logging(logging.DEBUG)

    parser = argparse.ArgumentParser(
        description="Detect well known errors",
        epilog="""
This script processes all log files against defined "known_problem" files,
caching the problems found, by package, into ".kpr" files.
""",
    )

    parser.add_argument(
        "sections",
        nargs="*",
        metavar="SECTION",
        help="limit processing to the listed SECTION(s)",
    )

    parser.add_argument(
        "--recheck",
        dest="recheck",
        action="store_true",
        help="recheck all log files (delete cache)",
    )

    parser.add_argument(
        "--recheck-failed",
        dest="recheck_failed",
        action="store_true",
        help="recheck failed log files (delete cache)",
    )

    args = parser.parse_args()

    conf = WKE_Config()
    conf.read(CONFIG_FILE)

    with open(os.path.join(conf["master-directory"], "dwke.lock"), "w") as lock:
        try:
            fcntl.flock(lock, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except IOError:
            if sys.stdout.isatty():
                sys.exit("another detect_well_known_errors process is already running")
            else:
                sys.exit(0)

        sections = args.sections
        if not sections:
            sections = conf["sections"].split()

        problem_list = create_problem_list(conf["known-problem-directory"])

        detect_well_known_errors(sections, conf, problem_list, args.recheck, args.recheck_failed)

# vi:set et ts=4 sw=4 :



# -*- coding: utf-8 -*-

# Copyright 2005 Lars Wirzenius (liw@iki.fi)
# Copyright © 2012-2013 Andreas Beckmann (anbe@debian.org)
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <https://www.gnu.org/licenses/>


#
# NOTA BENE: This module MUST NOT use the logging module, since it can and
# will be used before logging is set up (because logging can be configured
# in a configuration file, that's why).
#

import configparser
import re
import subprocess
from collections import UserDict, defaultdict
from functools import reduce

import distro_info


class MissingSection(Exception):
    def __init__(self, filename, section):
        self.args = ("Section %s not defined in configuration file %s" % (section, filename),)


class MissingMandatorySetting(Exception):
    def __init__(self, filename, key):
        self.args = ("Value for %s not set in configuration file %s" % (key, filename),)


class Config(UserDict):
    def __init__(self, section, defaults, mandatory=[], defaults_section=None):
        UserDict.__init__(self)
        self._section = section
        self._defaults_section = defaults_section
        for key, value in defaults.items():
            self[key] = value
        self._mandatory = mandatory

    def read(self, filename):
        cp = configparser.ConfigParser()
        cp.read(filename)
        if not cp.has_section(self._section):
            raise MissingSection(filename, self._section)
        for key in self.keys():
            if cp.has_option(self._section, key):
                self[key] = cp.get(self._section, key)
            elif self._defaults_section and cp.has_option(self._defaults_section, key):
                self[key] = cp.get(self._defaults_section, key)
            elif key in self._mandatory:
                raise MissingMandatorySetting(filename, key)

    def get_mirror(self, distro=None):
        if self["mirror"] is not None:
            return self["mirror"]
        return "http://deb.debian.org/debian"

    def get_distros(self):
        if self["upgrade-test-distros"] is not None:
            return self["upgrade-test-distros"].split()
        return []

    def get_distro(self):
        if self["distro"]:
            return self["distro"]
        distros = self.get_distros()
        if distros:
            return distros[-1]
        return None

    def get_start_distro(self):
        distros = self.get_distros()
        if distros:
            return distros[0]
        return self["distro"]

    def get_final_distro(self):
        distros = self.get_distros()
        if distros:
            return distros[-1]
        return self["distro"]

    def _get_distmap(self):
        debdist = distro_info.DebianDistroInfo()

        # start with e.g. "sid" -> "unstable"
        distmap = defaultdict(
            lambda: "unknown",
            [
                (debdist.old(), "oldstable"),
                (debdist.devel(), "unstable"),
                (debdist.stable(), "stable"),
                (debdist.testing(), "testing"),
                ("experimental", "experimental"),
                ("rc", "experimental"),
            ],
        )

        # add mappings for e.g. "oldstable" -> "oldstable"
        distmap.update({val: val for val in distmap.values()})

        # map e.g. "Debian6" -> "oldstable" where debdist.old(result="fullname")
        # currently returns 'Debian 6.0 "Squeeze"'
        dfuncs = [debdist.old, debdist.stable, debdist.testing]
        distmap.update({("Debian" + re.split("[ .]", x(result="fullname"))[1]): distmap[x()] for x in dfuncs})

        return distmap

    def _map_distro(self, distro):
        distro_root = re.split("[-.]", distro)[0]
        distmap = self._get_distmap()
        return distmap[distro_root]

    def get_std_distro(self, distrolist=[]):
        if not distrolist:
            distrolist = [self.get_distro()] + self.get_distros()
        mappedlist = [self._map_distro(x) for x in distrolist]
        return reduce(lambda x, y: y if y != "unknown" else x, mappedlist)

    def get_area(self):
        if self["area"] is not None:
            return self["area"]
        return "main"

    def get_arch(self):
        if not self["arch"]:
            # Try to figure it out ourselves, using dpkg
            p = subprocess.Popen(["dpkg", "--print-architecture"], stdout=subprocess.PIPE)
            self["arch"] = p.stdout.read().decode().rstrip()
        return self["arch"]


class DistroConfig(UserDict):
    def __init__(self, filename, mirror):
        UserDict.__init__(self)
        self._mirror = mirror
        self._defaults = {
            "uri": None,
            "distribution": None,
            "components": None,
            "target-release": None,
            "depends": None,
            "candidates": None,
        }
        cp = configparser.ConfigParser()
        cp.read(filename)
        for section in cp.sections():
            self[section] = dict(self._defaults)
            for key in self._defaults.keys():
                if cp.has_option(section, key):
                    self[section][key] = cp.get(section, key)

    def get(self, section, key=None):
        if section not in self.keys():
            self[section] = dict(self._defaults, distribution=section)
        if key is not None:
            return self[section][key]
        return self[section]

    def _is_virtual(self, distro):
        uri = self.get(distro, "uri")
        return uri is not None and uri == "None"

    def get_mirror(self, distro):
        if self._is_virtual(distro):
            distro = self._expand_depends(distro)[0]
        return self.get(distro, "uri") or self._mirror

    def get_distribution(self, distro):
        if self._is_virtual(distro):
            distro = self._expand_depends(distro)[0]
        return self.get(distro, "distribution") or distro

    def get_candidates(self, distro):
        return (self.get(distro, "candidates") or "").split() or [distro]

    def _get_packages_url(self, distro, area, arch):
        return "%s/dists/%s/%s/binary-%s/Packages" % (
            self.get_mirror(distro),
            self.get_distribution(distro),
            area,
            arch,
        )

    def get_packages_urls(self, distro, area, arch):
        return [self._get_packages_url(d, area, arch) for d in self.get_candidates(distro)]

    def _get_sources_url(self, distro, area):
        return "%s/dists/%s/%s/source/Sources" % (
            self.get_mirror(distro),
            self.get_distribution(distro),
            area,
        )

    def get_sources_urls(self, distro, area):
        return [self._get_sources_url(d, area) for d in self.get_candidates(distro)]

    def get_target_flags(self, distro):
        tr = self.get(distro, "target-release")
        if tr:
            return ["-t", tr]
        return []

    def _expand_depends(self, distro, include_virtual=False):
        todo = [distro]
        done = []
        seen = []
        while todo:
            curr = todo[0]
            todo = todo[1:]
            if curr not in seen:
                seen.append(curr)
                todo = (self.get(curr, "depends") or "").split() + [curr] + todo
            elif curr not in done:
                if include_virtual or not self._is_virtual(curr):
                    done.append(curr)
        assert len(done) > 0
        return done

    def get_deb_lines(self, distro, components):
        lines = []
        for d in self._expand_depends(distro):
            for c in components:
                if self[d]["components"] is None or c in self[d]["components"].split():
                    lines.append("deb %s %s %s" % (self.get_mirror(d), self.get_distribution(d), c))
        return lines

    def get_basetgz(self, distro, arch, merged_usr=True):
        # look for the first base distribution
        for d in self._expand_depends(distro):
            if self.get(d, "depends"):
                next  # skip partial distro
            return "%s%s_%s.tar.gz" % (
                self.get_distribution(d),
                "-merged-usr" if merged_usr else "",
                arch,
            )
        return None


# vi:set et ts=4 sw=4 :



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright 2005 Lars Wirzenius (liw@iki.fi)
# Copyright 2011 Mika Pflüger (debian@mikapflueger.de)
# Copyright © 2012-2017 Andreas Beckmann (anbe@debian.org)
# Copyright © 2020 Nicolas Dandrimont (nicolas@dandrimont.eu)
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <https://www.gnu.org/licenses/>


"""Analyze failed piuparts logs and move them around when the errors are known.

This program looks at piuparts log files in ./fail, and queries the bts to find
out if bugs have been filed already. If so, it moves them to ./bugged.
It tries to detect if new versions of bugged packages are uploaded without solving
the bug and will then update the bts with the new found versions, and copy the
headers of the log in ./fail to the one in ./bugged and vice versa. It will then
move the failed log to ./bugged as well.

"""
from __future__ import print_function

import fcntl
import os
import re
import shutil
import sys
import time
import traceback
from collections import defaultdict, deque
from functools import cmp_to_key

import apt_pkg
import debianbts

import piupartslib.conf
from piupartslib.conf import MissingSection

CONFIG_FILE = "/etc/piuparts/piuparts.conf"

apt_pkg.init_system()

error_pattern = re.compile(r"(?<=\n).*error.*\n?", flags=re.IGNORECASE)
chroot_pattern = re.compile(r"tmp/tmp.*?'")

############################################################################


class PiupartsBTS:
    def __init__(self):
        self._fetched = False
        self._bugs = {}
        self._bugs_in_package = defaultdict(set)
        self._bugs_affecting_package = defaultdict(set)
        self._bug_versions = {}

        self._queries = 0

    def fetch_bug_data(self):
        if self._fetched:
            return
        self._fetched = True
        bug_nums = debianbts.get_usertag("debian-qa@lists.debian.org", ["piuparts"])["piuparts"]
        bugs = debianbts.get_status(bug_nums)
        for bug in bugs:
            bug_num = bug.bug_num

            self._bugs[bug_num] = bug

            # Populate bug number -> versions map
            versions = []
            for version in bug.found_versions:
                # debianbts returns found versions in the format package/1.2.3 or 1.2.3 which will become 1.2.3
                v = version.rsplit("/", 1)[-1]
                if v == "None":
                    continue
                versions.append(v)
            if versions:
                versions.sort(key=cmp_to_key(apt_pkg.version_compare))
                self._bug_versions[bug_num] = versions

            # Populate package -> bug map
            for package in bug.package.split(","):
                if package.startswith("src:"):
                    package = package[4:]
                self._bugs_in_package[package].add(bug_num)

            # Populate affected package -> bug map
            for package in bug.affects:
                if package.startswith("src:"):
                    package = package[4:]
                self._bugs_affecting_package[package].add(bug_num)

    def bugs_in(self, package):
        self._queries += 1
        self.fetch_bug_data()
        return self._bugs_in_package[package]

    def bugs_affecting(self, package):
        self._queries += 1
        self.fetch_bug_data()
        return self._bugs_affecting_package[package]

    def bug_versions(self, bug):
        """Gets a list of only the version numbers for which the bug is found.
        Newest versions are returned first."""
        self._queries += 1
        self.fetch_bug_data()
        return self._bug_versions.get(bug, ["~"])

    def print_stats(self):
        print("PiupartsBTS: %d queries" % self._queries)


piupartsbts = PiupartsBTS()

############################################################################


class Busy(Exception):
    def __init__(self):
        self.args = ("section is locked by another process",)


class Config(piupartslib.conf.Config):
    def __init__(self, section="global", defaults_section=None):
        self.section = section
        piupartslib.conf.Config.__init__(
            self,
            section,
            {
                "sections": "report",
                "master-directory": ".",
            },
            defaults_section=defaults_section,
        )


def find_logs(directory):
    """Returns list of logfiles sorted by age, newest first."""
    logs = [os.path.join(directory, x) for x in os.listdir(directory) if x.endswith(".log")]
    return [y[1] for y in reversed(sorted([(os.path.getmtime(x), x) for x in logs]))]


def find_bugged_logs(failed_log):
    package = package_name(failed_log)
    pat = "/" + package + "_"
    return [x for x in find_logs("bugged") + find_logs("affected") if pat in x]


def package_name(log):
    return os.path.basename(log).split("_", 1)[0]


def package_version(log):
    return os.path.basename(log).split("_", 1)[1].rstrip(".log")


def package_source_version(log):
    version = package_version(log)
    possible_binnmu_part = version.rsplit("+", 1)[-1]
    if possible_binnmu_part.startswith("b") and possible_binnmu_part[1:].isdigit():
        # the package version contains a binnmu-part which is not part of the source version
        # and therefore not accepted/tracked by the bts. Remove it.
        version = version.rsplit("+", 1)[0]
    return version


def extract_errors(log):
    """This pretty stupid implementation is basically just 'grep -i error', and then
    removing the timestamps and the name of the chroot and the package version itself.
    """
    with open(log, "r") as f:
        data = f.read()
    whole = ""
    pversion = package_version(log)
    for match in error_pattern.finditer(data):
        text = match.group()
        # Get rid of timestamps
        if text[:1].isdigit():
            text = text.split(" ", 1)[1]
        # Get rid of chroot names
        if "tmp/tmp" in text:
            text = re.sub(chroot_pattern, "chroot'", text)
        # Get rid of the package version
        text = text.replace(pversion, "")
        whole += text
    return whole


def extract_headers(log):
    with open(log, "r") as f:
        data = f.read()
    headers = []
    headers = data.partition("\nExecuting:")[0]
    if headers and not headers.endswith("\n"):
        headers += "\n"
    return headers


def prepend_to_file(filename, data):
    with open(filename, "r") as f:
        old_data = f.read()
    with open(filename + ".tmp", "w") as f:
        f.write(data)
        f.write(old_data)

    shutil.copymode(filename, filename + ".tmp")

    os.rename(filename, filename + "~")
    os.rename(filename + ".tmp", filename)
    os.remove(filename + "~")


def write_bug_file(failed_log, bugs):
    if bugs:
        with open(os.path.splitext(failed_log)[0] + ".bug", "w") as f:
            for bug in bugs:
                f.write('<a href="https://bugs.debian.org/%s" target="_blank">#%s</a>\n' % (bug, bug))


def move_to_bugged(failed_log, bugged="bugged", bug=None):
    print("Moving %s to %s (#%s)" % (failed_log, bugged, bug))
    os.rename(failed_log, os.path.join(bugged, os.path.basename(failed_log)))
    if bug is not None:
        write_bug_file(os.path.join(bugged, os.path.basename(failed_log)), [bug])


def mark_bugged_version(failed_log, bugged_log):
    """Copies the headers from the old log to the new log and vice versa and
    moves the new log to bugged. Removes the old log in bugged."""
    bugged_headers = extract_headers(bugged_log)
    failed_headers = extract_headers(failed_log)
    prepend_to_file(failed_log, bugged_headers)
    prepend_to_file(bugged_log, failed_headers)
    move_to_bugged(failed_log)


def bts_update_found(bugnr, newversion):
    if "DEBEMAIL" in os.environ and os.environ["DEBEMAIL"]:
        # subprocess.check_call(('bts', 'found', bugnr, newversion))
        print(" ".join(("bts", "found", str(bugnr), newversion)))


def mark_logs_with_reported_bugs():
    for failed_log in find_logs("fail") + find_logs("untestable"):
        try:
            pname = package_name(failed_log)
            pversion = package_source_version(failed_log)
            try:
                failed_errors = extract_errors(failed_log)
            except IOError:
                print("IOError while processing %s" % failed_log)
                continue
            moved = False
            abugs = piupartsbts.bugs_affecting(pname)
            bugs = piupartsbts.bugs_in(pname)
            all_bugs = sorted(abugs | bugs, reverse=True)
            for bug in all_bugs:
                if moved:
                    break
                if bug in abugs:
                    bugged = "affected"
                else:
                    bugged = "bugged"
                found_versions = piupartsbts.bug_versions(bug)
                if pversion in found_versions:
                    move_to_bugged(failed_log, bugged, bug)
                    moved = True
                    break
                for bug_version in found_versions:
                    # print('DEBUG: %s/%s #%d %s' % (pname, pversion, bug, bug_version))

                    if apt_pkg.version_compare(pversion, bug_version) > 0:  # pversion > bug_version
                        bugged_logs = find_bugged_logs(failed_log)
                        if not bugged_logs and not moved:
                            print(
                                "%s/%s: Maybe the bug was filed earlier: https://bugs.debian.org/%d against %s/%s"
                                % (pname, pversion, bug, pname, bug_version)
                            )
                            break
                        for bugged_log in bugged_logs:
                            old_pversion = package_source_version(bugged_log)
                            bugged_errors = extract_errors(bugged_log)
                            if (
                                apt_pkg.version_compare(old_pversion, bug_version) == 0  # old_pversion == bug_version
                                and failed_errors == bugged_errors
                            ):
                                # a bug was filed for an old version of the package,
                                # and the errors were the same back then - assume it is the same bug.
                                if not moved:
                                    mark_bugged_version(failed_log, bugged_log)
                                    moved = True
                                    bts_update_found(bug, pversion)
                                    break
            if not moved:
                write_bug_file(failed_log, all_bugs)
        except KeyboardInterrupt:
            raise
        except Exception:
            print("ERROR processing %s" % failed_log)
            traceback.print_exc()


def main():
    conf = Config()
    conf.read(CONFIG_FILE)

    master_directory = conf["master-directory"]
    if len(sys.argv) > 1:
        sections = sys.argv[1:]
    else:
        sections = conf["sections"].split()

    with open(os.path.join(master_directory, "analyze.lock"), "w") as lock:
        try:
            fcntl.flock(lock, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except IOError:
            if sys.stdout.isatty():
                sys.exit("another piuparts-analyze process is already running")
            else:
                sys.exit(0)

        todo = deque([(s, 0) for s in sections])
        while len(todo):
            (section_name, next_try) = todo.popleft()
            now = time.time()
            if now < next_try:
                print("Sleeping while section is busy")
                time.sleep(max(30, next_try - now) + 30)
            print(time.strftime("%a %b %2d %H:%M:%S %Z %Y", time.localtime()))
            print("%s:" % section_name)
            try:
                section_directory = os.path.join(master_directory, section_name)
                if not os.path.exists(section_directory):
                    raise MissingSection("", section_name)
                with open(os.path.join(section_directory, "master.lock"), "w") as lock:
                    try:
                        fcntl.flock(lock, fcntl.LOCK_EX | fcntl.LOCK_NB)
                    except IOError:
                        raise Busy()

                    oldcwd = os.getcwd()
                    os.chdir(section_directory)
                    mark_logs_with_reported_bugs()
                    os.chdir(oldcwd)
            except Busy:
                print("Section is busy")
                todo.append((section_name, time.time() + 60))
            except MissingSection as e:
                print("Configuration Error in section '%s': %s" % (section_name, e))
            print("")

        print(time.strftime("%a %b %2d %H:%M:%S %Z %Y", time.localtime()))
        piupartsbts.print_stats()


if __name__ == "__main__":
    main()

# vi:set et ts=4 sw=4 :
